/**
 * description
 *
 * @author eson
 *2022年6月15日-17:38:16
 */
package com.yuandian.common;

import java.io.IOException;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.Executor;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.function.Function;
import org.yaml.snakeyaml.Yaml;
import com.alibaba.nacos.api.NacosFactory;
import com.alibaba.nacos.api.PropertyKeyConst;
import com.alibaba.nacos.api.config.ConfigService;
import com.alibaba.nacos.api.config.listener.Listener;
import com.alibaba.nacos.api.exception.NacosException;
import lombok.Getter;
import lombok.Setter;
import lombok.extern.slf4j.Slf4j;

/**
 * description nacos配置.
 *
 * @author eson
 *         2022年6月13日-17:08:46
 */
@Slf4j
@Getter
@Setter
public class Config {

	/** 加载的配置文件名, 在{@docRoot}/resources下 */
	public static String DEFAULT_CONFIT_FILE = "application.properties";
	// 默认
	public static String DEFAULT_GROUP_DATAID = "yuandian.dataflow";
	// 所有生成的nacos客户端
	private static HashMap<String, Config> configDict = new HashMap<>();

	/** properties 的配置key. nacos地址 */
	private String DEFAULT_CONFIG_ADDR = "yuandian.dataflow.config.nacos.server.addr";
	// 配置的所有值主类
	public Map<String, Object> data;
	// nacos地址
	public String serverAddr;
	// nacos dataId
	public String dataId;
	// nacos group
	public String group;
	// 线程安全配置锁
	private Lock datalock;
	// nacos 客户端类
	private ConfigService configService;

	private Config(String GroupAndDataId) throws Exception {

		String[] gad = GroupAndDataId.split("\\.");
		if (gad.length != 2) {
			throw new Exception("Group 或者 DataId 不能存在 '.' 的命令");
		}
		this.group = gad[0];
		this.dataId = gad[1];

		if(ENV_TEST != null) {
			DEFAULT_CONFIG_ADDR += ENV_TEST;
			this.group += ENV_TEST;
			this.dataId += ENV_TEST;
		}

		connect();
	}

	/**
	 * 连接nacos
	 * @throws Exception
	 */
	private void connect() throws Exception {
		if (configService != null) {
			configService.shutDown();
			configService = null;
		}
		// 获取 app
		Properties prop = new Properties();
		prop.load(this.getClass().getClassLoader().getResourceAsStream(DEFAULT_CONFIT_FILE));

		serverAddr = trim(prop.getProperty(DEFAULT_CONFIG_ADDR), "\" '");
 
		Properties properties = new Properties();
		
		properties.put(PropertyKeyConst.SERVER_ADDR, serverAddr);
		configService = NacosFactory.createConfigService(properties);
 
		String content = configService.getConfig(dataId, group, 5000);
		if(content == null) {
			throw new Exception( String.format("content is null. maybe timeout or the error of addr %s", serverAddr));
		}
	 
		data = new Yaml().load(content);
		log.debug(content);

		datalock = new ReentrantLock();
		// 监听 配置更新事件
		configService.addListener(dataId, group, new Listener() {
			@Override
			public void receiveConfigInfo(String configInfo) {
				// log.debug("recieve:" + configInfo);
				try {
					datalock.lock();
					data = (Map<String, Object>) new Yaml().load(configInfo);
					// log.debug("{}", data);
				} finally {
					datalock.unlock();
				}
			}

			@Override
			public Executor getExecutor() {
				return null;
			}
		});
	}

	/**
	 * 根据多个key获取yaml的值 keys 路径
	 * 
	 * @param keys 获取的key值
	 * @return
	 */
	public Object get(String... keys) {
		var cur = data;
		for (var i = 0; i < keys.length - 1; i++) {
			var key = keys[i];
			cur = (Map<String, Object>) cur.get(key);
		}
		return cur.get(keys[keys.length - 1]);
	}

	/**
	 * 
	 * 用于定位keys的路径后的操作. 创建keys后赋值. 如果存在keys, 可以直接赋值, 不存在则直接报错
	 *
	 * @author eson
	 *         2022年6月15日-下午12:06:37
	 */
	public class Operator {
		Config config;
		String[] keys;

		Operator(Config config, String[] keys) {
			this.config = config;
			this.keys = keys;
		}

		/**
		 * 创建seek的key
		 * 
		 * @return 返回自身。
		 */
		public Operator createKeys() {
			var cur = config.data;
			for (var i = 0;;) {
				var key = keys[i];
				var vobj = cur.get(key);
				if (vobj == null) {
					vobj = new LinkedHashMap<>();
					cur.put(key, vobj);
				}
				i++;
				if (i >= keys.length) {
					break;
				}
				cur = (Map<String, Object>) vobj;
			}
			return this;
		}

		/**
		 * 定位后赋值
		 * 
		 * @param value 赋值
		 */
		public void set(Object value) {
			var cur = config.data;
			for (var i = 0; i < keys.length - 1; i++) {
				var key = keys[i];
				cur = (Map<String, Object>) cur.get(key);
			}
			cur.put(keys[keys.length - 1], value);
		}
	}

	/**
	 * 定位. eg. seek("key1", "key2")
	 * 
	 * @param keys
	 * @return 返回自身
	 */
	public Operator seek(String... keys) {
		return new Operator(this, keys);
	}

	/**
	 * 删除 key的值. 类型map操作. keys是一个路径 remove("a","b") --> {"a": {"b": 1}} 删除b
	 * 
	 * @param map
	 * @param keys
	 * @return
	 */
	public Object remove(String... keys) {
		var cur = data;
		for (var i = 0; i < keys.length - 1; i++) {
			var key = keys[i];
			cur = (Map<String, Object>) cur.get(key);
		}
		return cur.remove(keys[keys.length - 1]);
	}

	/**
	 * 更新配置
	 * 
	 * @return 返回是否发布成功.
	 * @throws NacosException
	 */
	public Boolean update() throws NacosException {
		return configService.publishConfig(dataId, group, new Yaml().dumpAsMap(data));
	}

	public static String ENV_TEST = null;

	/**
	 * 统一使用配置的入口函数 线程安全
	 * 
	 * @param GroupAndDataID 使用配置的标签 eg."group.dataId"
	 * @param execute        匿名函数
	 * @return 函数返回的值, 如果不需要直接返回null
	 * @throws Exception
	 */
	public static Object UseConfig(String GroupAndDataID, Function<Config, Object> execute) throws Exception {
		log.info(GroupAndDataID);
		Config cnf;
		synchronized (configDict) {
			cnf = configDict.get(GroupAndDataID);
			if (cnf == null) {
				cnf = new Config(GroupAndDataID);
				configDict.put(GroupAndDataID, cnf);
			}
		}
		try {
			cnf.datalock.lock();
			var res = execute.apply(cnf);
			return res;
		} catch (Exception e) {
			throw e;
		} finally {
			cnf.datalock.unlock();
		}
	}

	/**
	 * 统一使用配置的入口函数 线程安全
	 * 
	 * @param execute 统一使用配置的入口匿名方法
	 * @return 函数返回的值, 如果不需要直接返回null
	 * @throws Exception
	 */
	public static Object UseConfig(Function<Config, Object> execute) throws Exception {
		return UseConfig(DEFAULT_GROUP_DATAID, execute);
	}

	/**
	 * 函数移除字符串两侧 预定义字符
	 * 
	 * @param trimedstr 需要修改的字符串
	 * @param mychars   定义修改的字符集合
	 * @return 修正后的字符串.
	 */
	private static String trim(String trimedstr, String mychars) {
		if (trimedstr == null) {
			return null;
		} else if (trimedstr.length() == 0) {
			return trimedstr;
		}

		var buf = trimedstr.toCharArray();
		var chars = mychars.toCharArray();
		var start = 0;

		TOPLOOP1: for (; start < buf.length; start++) {
			for (var c : chars) {
				if (buf[start] == c) {
					continue TOPLOOP1;
				}
			}
			break;
		}

		var end = buf.length - 1;
		TOPLOOP2: for (; end >= 0; end--) {
			for (var c : chars) {
				if (buf[end] == c) {
					continue TOPLOOP2;
				}
			}
			break;
		}
		return trimedstr.substring(start, end + 1);
	}


	public static void main(String[] args) throws Exception {
		try {
			Config.UseConfig((cnf)->{
				log.info("{}", cnf);
				return null;
			});
		} catch (Exception e) {
			//TODO: handle exception
			e.printStackTrace();
		}

	}
}